// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//--------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//--------------------------------------------------------------------------
//
//	SelectBlockFuncs.cc: --- operator functions which require reading a file
//
//	DESCRIPTION:
//		Implementation of file based operator functions:
//			gti(<gtifilename>, timecol)
//			mask(<maskfilename>, xoff, yoff, xcol, ycol)
//		For details see expanded description of evselect.
//
//	EXTERNAL DEPENDENCIES:
//		function			package				comments
//		----------------------------------------------------------------------
//		many DAL routines	SAS					some are not implemented
//
//	CHANGE HISTORY:
//		UL, 1998-01-05: 1.0.0: - first version
//		UL, 1998-04-10: 1.1.0: - filtering with GTI files improved
//		UL, 1998-05-19: 1.2.0: - various minor modifications to bring code
//							     in line with other changes in selectlib 2.0
//		UL, 1998-10-07: 1.3.0: - mask filter supports now WCS; this requires
//								 WCS keywords to be present in the mask
//								 image (CRPIX1/1, CRVAL1/2, CRDELT1/2)
//								 and the currently processed event list
//								 file (TCRPXx, TCRVLx, TCRDLTx)
//		UL, 1999-01-19: 1.4.0: - fix in MaskFilter::MaskFilter to allow
//								 mask filter in primary array
//		UL, 1999-01-20: 1.5.0: - bug fix in passFilter() which caused
//								 mask image to be applied Y-flipped
//		UL, 1999-03-03: 1.6.0: - changes enforced by Lowmem Dal
//		UL, 1999-11-02: 1.7.0: - added support for the new unsigned types
//								 in dal>=1.80
//		UL, 1999-11-17: 1.8.0: - brought in line with SetTableColumn 2.0.0
//		UL, 1999-12-17: 1.9.0: - GTI filter: issue warning message when
//								 an interval with start>=stop is seen
//							   - Region -> GeoRegion
//		UL, 2000-02-03: 1.10.0:- fixed SPR 101: Behaviour with empty GTI tables
//								 empty GTI tables were erroneoulsy treated as
//								 _open_ filters
//		UL, 2000-07-05: 1.10.1:- bug fix in MaskFilter::MaskFilter()
//		UL, 2000-10-25: 1.11.1:- changed text of warning message which gets
//								 issued when an empty GTI table is seen
//		UL, 2000-11-10: 1.11.2:- trivial change to make optimizer happy
//
//	AUTHOR:
//		U. Lammers (UL), Wed Aug 27 11:48:12 MET DST 1997
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/selcore/SelectBlockFuncs.cc,v 1.3 2009/06/26 15:51:59 aibarra Exp $
//

#include "basicdefs.h"
#include "../utils/SelectUtils.h"
#include "SelectBlockFuncs.h"
#include "RegionFilter.h"
#include "SelectError.h"
#include "TableDescriptor.h"
#include "Evaluator.h"

#include <vector>

//
//	----------------------------( GTI filter )------------------------------
//
/* AI gcc4 porting
 * static const  (type qualifiers ignored on function return type) => const removed
 */
  OpNode singleGti(const Gti &gti, const OpNode timeNode)
{
	if (gti.start() > gti.stop()) {
		selerrstr<<"A GTI with start time ("<<gti.start()<<") > stop "
				   "time ("<<gti.stop()<<") has been found - no event will "
				   "pass this filter"<<selwarning("DegenerateGTI");
		return new FalseNode;
	}

	OpNode gtNode = new Ge(timeNode, new NumericLiteral(gti.start())),
		   ltNode = new Lt(timeNode, new NumericLiteral(gti.stop()));

	return new AndIf(gtNode, ltNode);
} // singleGti()

/* AI gcc4 porting
 * static const  (type qualifiers ignored on function return type) => const removed
 */
  OpNode gtiTree(const GtiList &gtiList, unsigned nlo, unsigned nhi,
							OpNode timeNode)
{
	if (nlo < nhi) {
		//	the normal case
		unsigned nmid = (nlo+nhi)/2+1;
		OpNode expr = new Lt(timeNode,
							 new NumericLiteral(gtiList[nmid].start()));
		return new Ifthenelse(expr,
							  gtiTree(gtiList, nlo, nmid-1, timeNode),
							  gtiTree(gtiList, nmid, nhi, timeNode));
	} else if (nlo == nhi)
		//	a single GTI
		return singleGti(gtiList[nlo], timeNode);

	return new FalseNode;
}  // gtiTree()


 OpNode GtiFilter::expandGtis(const std::string &fileName,
								   OpNode timeNode)
{
	InOGIPgtiTable	inOGIPgtiTable(fileName);
	GtiList	gtiList;

	//	read the list of GTIs and reduce it
	inOGIPgtiTable >> gtiList;
	GtiList reducedList = gtiList.reduce();

	if (reducedList.size())
		return gtiTree(reducedList, 0, reducedList.size()-1, timeNode);

	selerrstr<<"The GTI table in `"<<fileName<<"' is empty - nothing will "
			   "pass this filter"<<selwarning("EmptyGtiTable");

	OpNode op = new FalseNode;
	return op;	//	optimizer does not like 'return new FalseNode;'
} // GtiFilter::expandGtis()


//
//	----------------------------( region filter )--------------------------
//
static OpNode singleElement(const FShape *element, const OpNode xNode,
								  const OpNode yNode)
{
	OpNode elementNode = 0;
	double xref = DBL(element->refPoint().x()),
		   yref = DBL(element->refPoint().y());

	switch (element->type()) {
		case ShapeBasics::POINT:	// this is just a point
			elementNode = new Fpoint(new NumericLiteral(xref),
									 new NumericLiteral(yref),
									 xNode, yNode);
			break;

		case ShapeBasics::SECTOR:
			elementNode = new Fsector(
							new NumericLiteral(xref), new NumericLiteral(yref),
							new NumericLiteral(DBL(element->rotAngles()[0])),
							new NumericLiteral(DBL(element->rotAngles()[1])),
							xNode, yNode);
			break;

		case ShapeBasics::CIRCLE:
			elementNode = new Fcircle(
							new NumericLiteral(xref), new NumericLiteral(yref),
							new NumericLiteral(DBL(element->radii()[0])),
							xNode, yNode);
			break;

		case ShapeBasics::ANNULUS:
			elementNode = new Fring(
							new NumericLiteral(xref), new NumericLiteral(yref),
							new NumericLiteral(DBL(element->radii()[0])),
							new NumericLiteral(DBL(element->radii()[1])),
							xNode, yNode);
			break;

		case ShapeBasics::ELLIPSE:
			elementNode = new Fellipse(
							new NumericLiteral(xref), new NumericLiteral(yref),
							new NumericLiteral(DBL(element->radii()[0])),
							new NumericLiteral(DBL(element->radii()[1])),
							new NumericLiteral(DBL(element->rotAngles()[0])),
							xNode, yNode);
			break;

		case ShapeBasics::ELLIPTANNULUS:
			elementNode = new Frellipse(
							new NumericLiteral(xref), new NumericLiteral(yref),
							new NumericLiteral(DBL(element->radii()[0])),
							new NumericLiteral(DBL(element->radii()[1])),
							new NumericLiteral(DBL(element->radii()[2])),
							new NumericLiteral(DBL(element->radii()[3])),
							new NumericLiteral(DBL(element->rotAngles()[0])),
							new NumericLiteral(DBL(element->rotAngles()[1])),
							xNode, yNode);
			break;

		case ShapeBasics::BOX:
		case ShapeBasics::ROTBOX:
			elementNode = new Fbox(
							new NumericLiteral(xref), new NumericLiteral(yref),
							//	.5 because we need the half-width
							new NumericLiteral(.5 *DBL(element->radii()[0])),
							new NumericLiteral(.5 *DBL(element->radii()[1])),
							new NumericLiteral(element->type() == ShapeBasics::ROTBOX
						  ? DBL(element->rotAngles()[0]) : 0.),	xNode, yNode);
			break;

		case ShapeBasics::RECTANGLE:
		case ShapeBasics::ROTRECTANGLE:
			elementNode = new Frectangle(
							new NumericLiteral(xref), new NumericLiteral(yref),
							new NumericLiteral(DBL(element->topRight().x())),
							new NumericLiteral(DBL(element->topRight().y())),
				new NumericLiteral(element->type() == ShapeBasics::ROTRECTANGLE
						? DBL(element->rotAngles()[0]) : 0.), xNode, yNode);
			break;

		case ShapeBasics::RHOMBUS:
		case ShapeBasics::ROTRHOMBUS:
			elementNode = new Fdiamond(
							new NumericLiteral(xref), new NumericLiteral(yref),
							//	.5 because we need the half-width
							new NumericLiteral(.5 *DBL(element->radii()[0])),
							new NumericLiteral(.5 *DBL(element->radii()[1])),
				new NumericLiteral(element->type() == ShapeBasics::ROTRHOMBUS
						? DBL(element->rotAngles()[0]) : 0.), xNode, yNode);
			break;

		case ShapeBasics::POLYGON: {
			std::vector<float> x = element->pointsX(),
							   y = element->pointsY();
			NodePointList *nodePointList = new NodePointList;

			for (unsigned n=0; n<x.size(); ++n)
				nodePointList->addPoint(new NumericLiteral(DBL(x[n])),
										new NumericLiteral(DBL(y[n])));

			elementNode =
					PolygonScope::newPolyFilter(nodePointList, xNode, yNode);
			break;
		}

		default:
			selerrstr<<"Region element of unknown type "<<UNT(element->type())
					 <<" - will be ignored"<<selwarning("UnknownRegionType");
			elementNode = new TrueNode;
			break;
	}

	//	check if region is excluded; if yes, return not-node
	return element->included() ? elementNode : new Not(elementNode);
} // singleElement()

/* AI gcc4 porting
 * static const  (type qualifiers ignored on function return type) => const removed
 */
static OpNode componentTree(unsigned nElements,
								  const Component &components,
								  const OpNode subTree, const OpNode xNode,
								  const OpNode yNode)
{
	if (nElements > 0) {
		OpNode leftElement = singleElement(components[nElements-1],
										   xNode, yNode);
		if (subTree) {
			//	combine the two trees
			OpNode combinedElements = new AndIf(leftElement, subTree);

			return componentTree(nElements-1, components, combinedElements,
								 xNode, yNode);
		} else
			return componentTree(nElements-1, components, leftElement,
								 xNode, yNode);
	}

	return subTree;
} // componentTree()


static OpNode regionFilterTree(unsigned nComponents, const GeoRegion &region,
							   const OpNode subTree, const OpNode xNode,
							   const OpNode yNode)
{
	if (nComponents > 0) {
		OpNode leftComponent = componentTree(region[nComponents-1].size(),
										     region[nComponents-1], 0,
										     xNode, yNode);
		if (subTree) {
			OpNode combinedComponent = new OrIf(leftComponent, subTree);

			return regionFilterTree(nComponents-1, region, combinedComponent,
									xNode, yNode);
		} else
			return regionFilterTree(nComponents-1, region, leftComponent,
									xNode, yNode);
	}

	return subTree;
} // regionFilerTree()

/* AI gcc4 porting
 * const  (type qualifiers ignored on function return type) => const removed
 */
 OpNode RegionFilter::expandRegion(const std::string &fileName,
										const OpNode xNode, const OpNode yNode)
{
	InASCregionTable inRegionTable(fileName);
	GeoRegion region;

	inRegionTable >> region;	// read region filter from set

	if (region.size())
		return regionFilterTree(region.size(), region, 0, xNode, yNode);

	//	empty table - filtering has no effect
	selerrstr<<"The region table in `"<<fileName<<"' is empty - filtering "
			   "will have no effect"<<selwarning("EmptyRegionTable");

	return new TrueNode;
} // RegionFilter::expandRegion()

 /* AI gcc4 porting
  * static const  (type qualifiers ignored on function return type) => const removed
  */
 OpNode RegionFilter::expandRegion(const std::string &fileName)
{
	InASCregionTable inRegionTable(fileName);
	GeoRegion region;

	inRegionTable >> region;	// read region filter from set

	if (region.size())
		return regionFilterTree(region.size(), region, 0,
			selectorGlobals->symbolTable->identifier(inRegionTable.xcoord()),
			selectorGlobals->symbolTable->identifier(inRegionTable.ycoord()));

	//	empty table - filtering has no effect
	selerrstr<<"RegionFilter::expandRegion: Warning: The region table in `"
			 <<fileName<<"' is empty - filtering will have no effect\n";

	return new TrueNode;
} // RegionFilter::expandRegion()


//
//	---------------------------( image mask filter )-----------------------
//
static bool passFilter(const OpNode me, unsigned xoff, unsigned yoff,
					   unsigned x, unsigned y)
{
	if (x < xoff || x >= me->auxData(0)+xoff ||
		y < yoff || y >= me->auxData(1)+yoff)
		return true;
	else {
		long index = (y-yoff)*me->auxData(0) + (x-xoff);
//		long index = (me->auxData(1)-(y-yoff)-1)*me->auxData(0) + (x-xoff);
//	Y-flipped - SSC-SPR 872
//

		if (index < 0 || index >= LNG(me->auxData(0) * me->auxData(1)))
			return true;

		return me->arrayData()[index] != 0;
	}
} // passFilter()


static bool maskFilterLogic(const OpNode me)
{
	return passFilter(me,
					  UNT(me->child(0)->arith(me->child(0))),	// xoffset
					  UNT(me->child(1)->arith(me->child(1))),	// yoffset
					  UNT(me->child(2)->arith(me->child(2))),	// x coordinate
					  UNT(me->child(3)->arith(me->child(3))));	// y coordinate
} // maskFilterLogic()


static double maskFilterLogicAsArith(const OpNode me)
{ return maskFilterLogic(me) ? 1. : 0.; }


static bool maskFilterLogicWCS(const OpNode me)
{
	double	 xWCS = (*me->eventTableWCS(0))
					(INT(me->child(2)->arith(me->child(2)))),// x in WCS
			 yWCS = (*me->eventTableWCS(1))
					(INT(me->child(3)->arith(me->child(3))));// y in WCS
	const SelectWCS *mx = me->maskWCS(0), *my = me->maskWCS(1);

	return passFilter(me,
					  UNT(me->child(0)->arith(me->child(0))),	// xoffset
					  UNT(me->child(1)->arith(me->child(1))),	// yoffset
					  UNT((xWCS - mx->refPixelValue())/mx->delta() +
						  mx->refPixel()),
					  UNT((yWCS - my->refPixelValue())/my->delta() +
						  my->refPixel()));
} // MaskFilterLogicWCS


static double maskFilterLogicAsArithWCS(const OpNode me)
{ return maskFilterLogicWCS(me) ? 1. : 0.; }


template<class ArrayT>
void fillArray(unsigned *arrayTo, ArrayT *arrayFrom, unsigned size)
{
	for (unsigned n=0; n<size; ++n)
		arrayTo[n] = UNT(arrayFrom[n]);
} // fillArray()


void MaskFilter::getWCSfromEventTable(const Table *table)
{
	if (_xcol.length() && _ycol.length()) {
		_eventTableWCS[0] =	SelectWCS(table, _xcol);
		_eventTableWCS[1] =	SelectWCS(table, _ycol);

		//	register evaluation function which consider WCS
		registerLogicFunc(maskFilterLogicWCS);
		registerArithFunc(maskFilterLogicAsArithWCS);
	}
} // getWCSfromEventTable()


MaskFilter::MaskFilter(const std::string &fileName,
					   const OpNode xoffset, const OpNode yoffset,
					   const OpNode xNode, const OpNode yNode)
{
	OpNode nd[] = { xoffset, yoffset, xNode, yNode };
	addChildren(NUM_OF(nd), nd);

	//	for now, register evaluation function which disregard WCS
	registerLogicFunc(maskFilterLogic);
	registerArithFunc(maskFilterLogicAsArith);

	//	read mask image from file and initialize 4th child with pointer
	//	to array data
	SetBlock sa(fileName);
	Array *array = sa.array();

	//	read the WCS for the mask
	_maskWCS[0] = SelectWCS(array, 1);
	_maskWCS[1] = SelectWCS(array, 2);
	if (_maskWCS[0].keywordsPresent() && _maskWCS[1].keywordsPresent()) {
		//	we have WCS information in the mask image - register function
		//	to read corresponding information from event list table
		if (xNode->isIdentifierNode())
			_xcol = xNode->name();
		if (yNode->isIdentifierNode())
			_ycol = yNode->name();

		if (selectorGlobals->tableDescriptor)
			getWCSfromEventTable(selectorGlobals->tableDescriptor->table());
	}

	std::vector<unsigned long> extents = array->size();

	//	bug in DAL - gives extents reversed
	_extent[0] = extents[1];
	_extent[1] = extents[0];

	unsigned long size = _extent[0] * _extent[1];
	_arrayData = new unsigned[size];		// memory for array data

	ArrayData *ad = array->data();
	switch (array->dataType()) {
		case Dal::Bool:
			fillArray(_arrayData, ad->boolData(), size); break;
		case Dal::Int8:
			fillArray(_arrayData, ad->int8Data(), size); break;
		case Dal::Int16:
			fillArray(_arrayData, ad->int16Data(), size); break;
		case Dal::Uint16:
			fillArray(_arrayData, ad->uint16Data(), size); break;
		case Dal::Int32:
			fillArray(_arrayData, ad->int32Data(), size); break;
		case Dal::Uint32:
			fillArray(_arrayData, ad->uint32Data(), size); break;
		default:
			selerrstr<<"The mask image "<<array->qualifiedName()
					 <<" passed to the mask filter is not "
					   "of integral (int8/16/32/uint16/32) or boolean type "
					   "- cannot proceed"<<selerror("NotIntegralType", 2);
			SELTHROW(SelectException);
			break;
	}

	//	tidy up
	delete ad;
	dataSetServer->close(sa.dataSet());
} // MaskFilter::MaskFilter()


MaskFilter::~MaskFilter()
{
	delete [] _arrayData;
} // MaskFilter::~MaskFilter()






