! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!------------------------------------------------------------------------------
!
!                       XMM Science Analysis System (SAS)
!                     (c) 1997-2010 European Space Agency
!
!------------------------------------------------------------------------------
!
!   arrayeval.f90 --- arithmetic with arrays
!
!   DESCRIPTION:
!       A module to perform arithmetic with arrays driven by an arithmetic
!       expression involving other array names, operators, functions, etc.
!
!   EXPORTED FUNCTIONS:
!       initArrayEval --- initialize the evaluator with arithmetic expression
!       doArrayEval   --- perform the operation
!
!   CHANGE HISTORY:
!       UL, 1997-07-25: 1.0.0: - first version
!
!   AUTHOR:
!       U. Lammers, Wed Mar 25 10:51:52 MET 1998
!
!   $Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/service/arrayeval.f90,v 1.2 2002/07/09 14:06:31 ulammers Exp $
!
module ArrayEval
    implicit none

    interface
        !
        !   SYNOPSIS:
        !      initArrayEval(expression)
        !
        !   PURPOSE:
        !      Initialize the array evaluation module with an arithmetic
        !      expression.
        !
        !   ARGUMENTS:
        !      In:
        !           expression: a string which must represent a valid
        !                       arithmetic expression otherwise an error
        !                       condition is raised; the expression can contain
        !                       numeric and symbolic (#E, #PI, etc.) constants,
        !                       operators (+, -, *, /, %, **, etc.), functions
        !                       (sin, cos, sqrt, ln, etc.) and symbolic names.
        !                       The latter represent names of arrays which are
        !                       to be passed in the subsequent call to the
        !                       doArrayEval subroutine. Array names must start
        !                       with a letter, optionally followed by a
        !                       characters sequence consisting of letters,
        !                       digits, and the underscore (_) character (like
        !                       variable names in F90/C++). Example:
        !
        !                            A1 + sin(A2) - exp(A1**2) / A3 + A2
        !
        subroutine initArrayEval(expression)
             character(len=*), intent(in)               :: expression
        end subroutine


        !
        !   SYNOPSIS:
        !      doArrayEval(inArrays, n, outArray)
        !
        !   PURPOSE:
        !      This subroutine does the array arithmetic based on the
        !      expression passed to initArrayEval above. An array of
        !      DAL array pointers must be given which must match in number
        !      and sequence the symbolic names used in the expression.
        !
        !   ARGUMENTS:
        !       In:
        !           inArrays:   a Fortran 90 array with pointers to Dal arrays
        !                       referred to in the expression via symbolic
        !                       names. For instance, if the above expression
        !                       was used to initialize the evaluator,
        !                       <inArray> must contain three pointers
        !                       to DAL arrays. The elements are associated
        !                       With the symbolic array names in the
        !                       expression in the sequence in which they
        !                       where used, e.g. inArray(0), inArray(1),
        !                       inArray(2) will be associated with A1, A2, and
        !                       A3 respectively. Upon successful exit, the
        !                       result of the evaluation will be placed in the
        !                       <outArray>. The input arrays need to have the
        !                       same dimensionality and extents but can differ\
        !                       in data type.
        !          n:           number of elements in <inArrays>
        !       Out:
        !           outArray:   output array to be populated; can be identical
        !                       to one of the input arrays in <inArrays>
        !
        subroutine doArrayEval(inArrays, n, outArray)
            use types
            use dal
            type(ArrayT), dimension(:), intent(in)     :: inArrays
            integer(kind=int32), intent(in)            :: n
            type(ArrayT), intent(inout)                :: outArray
        end subroutine
    end interface

end module ArrayEval





