// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------
//
//                      XMM Science Analysis System (SAS)
//                     (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	ExprTranslator.cc: --- implementation of ExprTranslator class
//
//	DESCRIPTION:
//		This implements the ExprTranlator class
//
//	CHANGE HISTORY:
//		UL, 1999-02-17: 1.0.0: - first version
//		UL, 1999-02-17: 1.1.0: - generated expression was invalid ;)
//		UL, 1999-09-20: 1.2.0: - take binning factors into account
//		UL, 1999-09-21: 1.3.0: - take also image offset into account
//		UL, 1999-10-25: 1.4.0: - added support for 2D regions from dsplot
//		UL, 1999-12-15: 1.4.1: - closed two minor memory leaks
//		UL, 2002-06-28: 1.5.0: - strstream->stringstream
//		UL, 2003-05-23: 1.6.0: - workaround for gcc compiler bug under MacOSX
//								 (search for 'MacOSX' below for more details)
//
//	AUTHOR:
//		U. Lammers (UL), Wed Feb 17 13:22:25 MET 1999
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/utils/ExprTranslator.cc,v 1.6 2009/06/26 15:51:59 aibarra Exp $
//
#include <iomanip>
#include <algorithm>
#include <sstream>

#include "common.h"
#include "stringutil.h"

#include "SelectError.h"
#include "ExprTranslator.h"

using namespace std;

string ImgdisplayRegionSpecification::toSelectlibExpression(
											const string &xcolName,
											const string &ycolName,
											double binFactorX,
											double binFactorY,
											double minX,
											double minY) const
{
	string region(_spec);	// make a copy
	char exclIncl = region[0];
	if (exclIncl != '+' && exclIncl != '-')
		throw InvalidRegionSpec();

	string::size_type pos;
	if ((pos = region.find("#")) != string::npos)
		region=trim(region.substr(0, pos-1));

	pos = region.find('(');
	string regionType(lowercase(region.substr(1, pos-1))),
		   regionArgs(region.substr(pos+1, region.find_last_of(')')-pos-1));

	if (regionType == "circle") {
		string rotAng = "";
		if (binFactorX != binFactorY) {
			regionType = "ellipse";	//	a circle becomes an ellipse for
									//	different bin factors
			rotAng = ",0";			//	0: rotation angle

		}
		regionArgs = circleArgs(regionArgs, binFactorX, binFactorY,
								minX, minY) + rotAng;
	} else if (regionType == "ellipse")
		regionArgs = ellipseArgs(regionArgs, binFactorX, binFactorY,
							     minX, minY);
	else if (regionType == "box")
		regionArgs = boxArgs(regionArgs, binFactorX, binFactorY, minX, minY);
	else if (regionType == "polygon")
		regionArgs = polygonArgs(regionArgs, binFactorX, binFactorY,
								 minX, minY);
	else if (regionType == "point")
		regionArgs = pointArgs(regionArgs, binFactorX, binFactorY, minX, minY);
	else
		errstr<<"`"<<regionType<<"' is not a supported region"
			  <<Sas::error("UnsupportedRegion");

	string expr('(' + xcolName + ',' + ycolName + ") IN " +
				regionType + '(' + regionArgs + ')');

	return exclIncl == '+' ? expr : "!" + expr;
} // ImgdisplayRegionSpecification::toSelectlibExpression()


string ImgdisplayRegionSpecification::circleArgs(const string &s,
					 double binX, double binY, double minX, double minY) const
{
	vector<string> v(::breakup(s, ","));
	if (v.size() != 3)
		errstr<<"Arguments of circle filter must be <x>,<y>,<r> but there "
				"is "<<s<<Sas::error("InputError");

	if (binX != binY)
		return convertPosition(v[0], binX, minX, v[1], binY, minY) + ',' +
			   convertDimension(v[2], binX) + ',' +
			   convertDimension(v[2], binY);

	string ret(convertPosition(v[0], binX, minX, v[1], binY, minY) + ',' +
			   convertDimension(v[2], binX));
	return ret;
} // ImgdisplayRegionSpecification::circleArgs()


string ImgdisplayRegionSpecification::ellipseArgs(const string &s,
					 double binX, double binY, double minX, double minY) const
{
	vector<string> v(::breakup(s, ","));
	if (v.size() != 5)
		errstr<<"Arguments of ellipse filter must be <x>,<y>,<r1>,<r2>, "
				"<ANGLE> but there is "<<s<<Sas::error("InputError");

	string ret(convertPosition(v[0], binX, minX, v[1], binY, minY) + ',' +
			   convertDimension(v[2], binX) + ',' +
			   convertDimension(v[3], binY) + ',' +
			   v[4]);
	return ret;
} // ImgdisplayRegionSpecification::ellipseArgs()


string ImgdisplayRegionSpecification::boxArgs(const string &s,
					 double binX, double binY, double minX, double minY) const
{
	vector<string> v(::breakup(s, ","));
	if (v.size() != 5)
		errstr<<"Arguments of box filter must be <x>,<y>,<w>,<y>,<angle> "
				"but there is "<<s<<Sas::error("InputError");


	string ret(convertPosition(v[0], binX, minX, v[1], binY, minY) + ',' +
			   convertDimension(v[2], .5 * binX) + ',' +
			   convertDimension(v[3], .5 * binY) + ',' +
			   v[4]);			//	selectlib box filter expects half-widths
	return ret;
} // ImgdisplayRegionSpecification::boxArgs()


string ImgdisplayRegionSpecification::polygonArgs(const string &s,
					 double binX, double binY, double minX, double minY) const
{
	vector<string> v(::breakup(s, ","));
	unsigned N = v.size();
	if (N % 2)
		errstr<<"Arguments of box filter must be <x1>,<y1>,<x2>,<y2>, "
				"..., <xn>,<yn> but there is "<<s<<Sas::error("InputError");

	string res;
	for (unsigned n=0; n<N/2; ++n)
		res += convertPosition(v[2*n], binX, minX, v[2*n+1], binY, minY) + ',';

	return res.substr(0, res.length()-1);
} // ImgdisplayRegionSpecification::polygonArgs()


string ImgdisplayRegionSpecification::pointArgs(const string &s,
					 double binX, double binY, double minX, double minY) const
{
	vector<string> v(::breakup(s, ","));
	if (v.size() != 2)
		errstr<<"Arguments of point filter must be <x>,<y> but"
				" there is "<<s<<Sas::error("InputError");

	return convertPosition(v[0], binX, minX, v[1], binY, minY);
} // ImgdisplayRegionSpecification::pointArgs()


string ImgdisplayRegionSpecification::convertPosition(
													  const string &x,
													  double binX, double minX,
													  const string &y,
													  double binY, double minY)
const
{
	double xv = toNumber<double>(x),
		   yv = toNumber<double>(y);

	return toString((xv-1.) * binX + minX) + ',' +
		   toString((yv-1.) * binY + minY);
} // ImgdisplayRegionSpecification::convertPosition()


string ImgdisplayRegionSpecification::convertDimension(
													   const string &dim,
													   double f) const
{
	return toString(toNumber<double>(dim) * f);
} // ImgdisplayRegionSpecification::convertDimension()

/* AI gcc4 porting
 * casting to (char*) needed
 */
DsplotRegionSpecification::RegionType
DsplotRegionSpecification::type(const vector<string> &comp) const
{
	static char *_s[] = {(char*)"above", (char*)"below", (char*)"left", (char*)"right", (char*)"polyi", (char*)"polyo" };
//
//	the following core dumps with gcc-3.2 under MacOSX for unknown reasons
//	- use copy() as workaround
//	static vector<string> s(_s, _s + NUM_OF(_s));
//
	vector<string> s;
	std::copy(_s, _s+NUM_OF(_s), back_inserter(s));

	RegionType reg = NONE;
	bool on = false;
	for (unsigned n=0; n<comp.size(); ++n) {
		vector<string> v(::breakup(comp[n], " ,"));
		if (v.size() < 1)
			continue;
		if (v[0] == "type") {
			vector<string>::const_iterator i =
			  find(s.begin(), s.end(), v[1]);
			if (i != s.end())
				reg = static_cast<RegionType>(i-s.begin());
			else
				throw InvalidRegionSpec();
		} else if (uppercase(v[0]) == "ON")
			on = true;
	}

	return on && reg != NONE ? reg : NONE;
} // DsplotRegionSpecification::type()

string DsplotRegionSpecification::toSelectlibExpression(
												const string &xColName,
												const string &yColName,
					 							double, double,
												double, double) const
{
	vector<string> comp(::breakup(_spec, ";"));
	if (comp.size() < 3)
		throw InvalidRegionSpec();

	RegionType whatType = type(comp);
	if (whatType == NONE)
		return "";

	//
	//	we have a polygon
	//
	vector<double> x, y;
	if (whatType == INSIDE_POLY || whatType == OUTSIDE_POLY) {
		for (unsigned n=1; n<comp.size(); ++n) {
			vector<string> s(::breakup(comp[n], " ,"));
			if (s.size() != 3)
				continue;
			if (s[0] != "xy")
				continue;
			x.push_back(toNumber<double>(s[1]));
			y.push_back(toNumber<double>(s[2]));
		}

		if (x.size() < 3)
			//	cannot have a polygon with less than 3 points
			throw InvalidRegionSpec();
	} else
		return "";

	stringstream s;
	if (whatType == OUTSIDE_POLY)
		//	outside polygon
		s<<'!';
	s<<'('<<xColName<<","<<yColName<<") IN polygon(";
	unsigned N = x.size();
	for (unsigned n=0; n<N; ++n) {
		s<<setprecision(6)<<x[n]<<","<<setprecision(6)<<y[n];
		if (n != N-1)
			s<<',';
	}
	s<<')';

	return s.str();
} // DsplotRegionSpecification::toSelectlibExpression()


string DsplotRegionSpecification::toSelectlibExpression(
										const string &colName) const
{
	vector<string> comp(::breakup(_spec, ";"));
	if (comp.size() < 3)
		throw InvalidRegionSpec();

	RegionType whatType = type(comp);

	unsigned n;
	double x1 = 0.,
		   y1 = 0.,
		   x2 = 0.,
		   y2 = 0.;
	for (n=0; n<comp.size(); ++n) {
		vector<string> v(::breakup(comp[n], " ,"));
		if (v[0] == "line") {
			//	found it
			x1 = toNumber<double>(v[1]);
			y1 = toNumber<double>(v[2]);
			x2 = toNumber<double>(v[3]);
			y2 = toNumber<double>(v[4]);
			break;
		}
	}

	if (n >= comp.size())
		throw InvalidRegionSpec();

	if (whatType != NONE) {
		//
		//	sometimes you get "line 0, 0, 0, 0"
		//

		if (whatType == ABOVE_LINE || whatType == BELOW_LINE) {
			//
			//	a single range
			//
			stringstream s1, s2;
			s1<<setprecision(6)<<(x1 < x2 ? x1 : x2);
			s2<<setprecision(6)<<(x1 < x2 ? x2 : x1);

			return colName + " IN [" + s1.str() + ":" + s2.str() + ']';
		} else if (whatType == LEFT_OF_LINE || whatType == RIGHT_OF_LINE) {
			//
			//	everything to the left or right of the specified line
			//	take the mean of x1/x2 as the value
			//
			stringstream s;
			s<<setprecision(6)<<.5*(x1+x2);

			return colName + (whatType == LEFT_OF_LINE ? "<=" : ">=") +s.str();
		} else
			throw InvalidType();
	}

	return "";
} // DsplotRegionSpecification::toSelectlibExpression()


