// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	Gti.cc:	--- implementation of <Gti> and <GtiList> classes
//
//	CHANGE HISTORY:
//		UL, 1998-02-24: 1.0.0: - first version
//		UL, 1999-??-??: 1.1.0: - updated
//		UL, 1999-10-21: 1.2.0: - added implementations of
//								 GtiList::operator&=/|=
//		UL, 2000-05-25: 1.3.0: - bug fix in GtiList::reduce()
//
//	AUTHOR:
//		U. Lammers (UL), Tue Feb 24 18:48:26 MET 1998
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/utils/Gti.cc,v 1.7 2010/02/25 17:14:07 aibarra Exp $
//

#include "Gti.h"
#include <numeric>
#include <algorithm>
/* AI gcc4 porting
 * error in operator os<<"("
 * We need to include sstream to deal with the ostream
 */
#include <sstream>

//AI. Initialization of static const (SOC-SPR-6595)
const double Gti::ETERNITY = 1.e99;

Gti::Gti(double start, double stop)
{
	if (start <= stop) {
		_start = start;
		_stop = stop;
	} else {
		_start = stop;
		_stop = start;
	}
} // Gti::Gti()


bool Gti::inside(double time) const
{
	return time >= _start && time <= _stop;
} // Gti::inside


std::ostream &operator<<(std::ostream &os, const Gti &gti)
{
	os<<"("<<gti.start()<<", "<<gti.stop()<<")";

	return os;
} // operator<<


bool operator<(const Gti &gti1, const Gti &gti2)
{
	if (gti1.start() != gti2.start())
		return gti1.start() < gti2.start();
	else
		return gti1.stop() < gti2.stop();
} // operator<


GtiList::GtiList(bool allGood) : _state(GtiList::UNKNOWN)
{
	if (allGood)
		this->push_back(Gti());
} // GtiList::GtiList()


const Gti *GtiList::inside(double time) const
{
	for (std::vector<Gti>::const_iterator i = this->begin();
		 i != this->end(); ++i)
		if (i->inside(time))
			return &(*i);

	return 0;
} // GtiList::inside


void GtiList::sortIt()
{
	if (_state != SORTED) {
		sort(this->begin(), this->end());
		_state = SORTED;
	}
} //	GtiList::sortIt()


GtiList &GtiList::reduce()
{
	if (_state == REDUCED)
		//	already reduced - nothing to do except returning list
		return *this;

	//	sort first
	sortIt();

	GtiList &gl = *this;
	GtiList newGtiList;
	newGtiList._state = REDUCED;
	unsigned nGtis = gl.size(),
			 nNewGtis = 0;
	if (nGtis == 0) {
		*this = newGtiList;
		return *this;
	}

	newGtiList.push_back(gl[0]);	//	copy first GTI
	for (unsigned n=1; n<nGtis; ++n)
		if (gl[n].start() > newGtiList[nNewGtis].stop()) {
			//
			//	a new interval
			//
			newGtiList.push_back(gl[n]);
			++nNewGtis;
		} else if (gl[n].stop() > newGtiList[nNewGtis].stop())
			//
			//	intervals overlap, and upper boundary of interval <n>
			//	extends over interval <nNewGtis>
			//
			newGtiList[nNewGtis].setStop(gl[n].stop());

	_state = REDUCED;
	*this = newGtiList;
	return *this;
} //	GtiList::reduce()


GtiList& GtiList::operator|=(const GtiList &list)
{
	unsigned N = list.size();
	this->reserve(this->size() + N);
	for (unsigned n=0; n<N; ++n)
		this->push_back(list[n]);
	_state = UNKNOWN;

	addOrigin(list.origins());

	return reduce();
} // GtiList::operator|=()


GtiList &GtiList::operator&=(const GtiList &l2)
{
	GtiList &l1 = *this;
	unsigned n1 = l1.size(),
			 n2 = l2.size();
	std::vector<std::string> origs(l1.origins());

	GtiList resList;
	resList.reserve(std::max(n1, n2));

	unsigned nRes = 0,
			 actn2 = 0;
	for (unsigned n=0; n<n1; ++n) {
		while (actn2<n2 && l2[actn2].stop() < l1[n].start())
			++actn2;
		while (actn2 < n2 && l2[actn2].start() <= l1[n].stop()) {
			//
			//	overlap found
			//
			resList.push_back(Gti(std::max(l1[n].start(), l2[actn2].start())));
			if (l1[n].stop() <= l2[actn2].stop()) {
				resList[nRes].setStop(l1[n].stop());
				++nRes;
				break;
			} else {
				resList[nRes].setStop(l2[actn2].stop());
				++nRes;
				++actn2;
			}
		}
		if (actn2 >= n2)
			break;
	}

	resList.resize(nRes);
	*this = resList;

	addOrigin(origs);
	addOrigin(l2.origins());

	return *this;
} // GtiList::operator&=()

namespace {
	double sumGti(double sum, const Gti &gti) { return sum + gti.length(); }
}

double GtiList::ontime() const
{
	return std::accumulate(begin(), end(), 0., sumGti);
} // GtiList::ontime()


std::ostream &operator<<(std::ostream &os, const GtiList &gtiList)
{
	for (std::vector<Gti>::const_iterator i = gtiList.begin();
		 i != gtiList.end(); ++i)
		os<<*i<<std::endl;

	return os;
} // operator<<


GtiList operator|(const GtiList &lhs, const GtiList &rhs)
{
	GtiList l(lhs);
	return l |= rhs;
} // operator|


GtiList operator&(const GtiList &lhs, const GtiList &rhs)
{
	GtiList l(lhs);
	return l &= rhs;
} // operator&=

